#
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            coordinates.py
#
# Author:          Olli Lammi
#
# Version:         0.8b
#
# Date:            05.08.2010
#
# Functions:       Translate
#                  KKJxy_to_WGS84lalo
#                  WGS84lalo_to_KKJxy
#                  KKJxy_to_KKJlalo
#                  KKJlalo_to_KKJxy
#                  KKJlalo_to_WGS84lalo
#                  WGS84lalo_to_KKJlalo
#                  KKJ_Zone_I
#                  KKJ_Zone_Lo
#                  KKJxy_ZoneShift
#                  WGS84lalo_to_GoogleMapsXY
#                  Str_to_CoordinateValue
#                  KKJxy_in_Finland
#                  ETRSTM35FINxy_to_WGS84lalo
#                  WGS84lalo_to_ETRSTM35FINxy
#                  KKJxy_to_ETRSTM35FINxy
#                  ETRSTM35FINxy_to_KKJxy
#                  WGS84distance
#                  WGS84bearing
#                   
# Description:     Coordinate system functions. Initial code and alhorithms
#                  extracted from Viestikallio's PHP-source code 
#                  (http://www.viestikallio.fi/tools/kkj-wgs84.php). 
#                  Python translation and changes by Olli Lammi.
#
#                  ETRS-TM35FIN algorithms implemented by Olli Lammi
#                  according to JHS 154.
#                  (http://docs.jhs-suositukset.fi/jhs-suositukset/JHS154/JHS154.pdf)
#
#                  Google Maps functions developed based on knowledge on
#                  multiple sites in the Internet, simplifying the algorithms
#                  and traditional trial error method.
# 
#                  NOTE!: The coordinate functions are developed to work only
#                  with coordinates that are in the area of Finland.
#
# Version history: ** 05.09.2005 v0.1a (Olli Lammi) **
#                  First version. Translated partially from PHP to Python.
#                  Original PHP code source: www.viestikallio.fi
# 
#                  ** 26.09.2005 v0.2a (Olli Lammi) **
#                  Included the WGS84_to_KKJxy -conversion.
#                  Changed all interfaces to work with degrees when
#                  using angle values
#                  Altered the KKJ zone info function to a
#                  lookup dictionary (KKJ_ZONE_INFO).
#                  Added function to calculate propable KKJ band
#                  from KKJ longitude (KKJ_Zone_Lo).
#                  
#                  ** 01.10.2005 v0.3a (Olli Lammi) **
#                  Small changes to function interfaces.
#
#                  ** 14.10.2005 v0.4a (Olli Lammi) **
#                  Added support for KKJ-bands 0 and 5.
#
#                  ** 26.01.2007 v0.5a (Olli Lammi) **
#                  Added prototype function to convert WGS84-coordinates
#                  to Google Maps URL tile parameter x and y.
#
#                  ** 19.02.2008 v0.6a (Olli Lammi) **
#                  Added funcion to convert WGS84 coordinate string to
#                  scalar coordinate value.
#
#                  ** 25.02.2008 v0.7a (Olli Lammi) **
#                  Added funcion to tell approximately whether given KKJ
#                  koordinate is in the are of Finland.
#
#                  ** 08.08.2008 v0.7b (Olli Lammi) **
#                  Added utility funcion to shift KKJxy coordinate to given 
#                  zone: KKJxy_ZoneShift
#
#                  ** 10.02.2010 v0.8a (Olli Lammi) **
#                  Added functions for ETRS-TM35FINxy to and from WGS84lalo 
#                  conversions. Added coordinate translate method according
#                  to type.
#
#                  ** 05.08.2010 v0.8b (Olli Lammi) **
#                  Increased iteration area for Lon value in KKJxy_to_KKJlalo 
#                  function. Added functions to calculate distance in meters
#                  and initial and final bearing between two WGS84 locations.
#
###########################################################################

# Imports

import sys, os, string
import math
import re


###########################################################################

# Constants

# Longitude0 and Center meridian of KKJ bands
KKJ_ZONE_INFO = { 0: (18.0,  500000.0), \
                  1: (21.0, 1500000.0), \
                  2: (24.0, 2500000.0), \
                  3: (27.0, 3500000.0), \
                  4: (30.0, 4500000.0), \
                  5: (33.0, 5500000.0), \
                }

# Coordinate system type identifiers
COORD_TYPE_YKJ = 'YKJxy'
COORD_TYPE_KKJ = 'KKJxy'
COORD_TYPE_WGS84 = 'WGS84lalo'
COORD_TYPE_ETRSTM35FIN = 'ETRSTM35FINxy'


# Functions

###########################################################################
# Function:  Translate
###########################################################################
# Input:     dictionary with ['type'] is coordinate system type identifier
#                            ['N'] is coordinate Northing / Lat
#                            ['E'] in coordinate Easting / Lon
#            type identifier of the coordinate system to transform the input
#                            coordinates to
# Output:    dictionary with ['type'] is coordinate system type identifier
#                            ['N'] is coordinate Northing / Lat
#                            ['E'] in coordinate Easting / Lon
###########################################################################

def Translate(coordIn, outType):
    if (coordIn['type'] == outType):
        return coordIn
    
    # first convert all formats to WGS84
    WGS = {}
    if (coordIn['type'] == COORD_TYPE_YKJ or coordIn['type'] == COORD_TYPE_KKJ):
        WGS = KKJxy_to_WGS84lalo({'P': coordIn['N'], 'I': coordIn['E']})
    elif (coordIn['type'] == COORD_TYPE_ETRSTM35FIN):
        WGS = ETRSTM35FINxy_to_WGS84lalo(coordIn)
    elif (coordIn['type'] == COORD_TYPE_WGS84):
        WGS['La'] = coordIn['N'] 
        WGS['Lo'] = coordIn['E']
    else:
        return None
       
    # then convert from WGS84 to wanted coordinate type 
    if (outType == COORD_TYPE_WGS84):
        return {'type': COORD_TYPE_WGS84, 'N': WGS['La'], 'E': WGS['Lo']}
    elif (outType == COORD_TYPE_KKJ or outType == COORD_TYPE_YKJ):
        KKJ = WGS84lalo_to_KKJxy(WGS)
        if (outType == COORD_TYPE_YKJ):
            KKJ = KKJxy_ZoneShift(KKJ, 3)
        return {'type': outType, 'N': KKJ['P'], 'E': KKJ['I']}
    elif (outType == COORD_TYPE_ETRSTM35FIN):
        ETRS = WGS84lalo_to_ETRSTM35FINxy(WGS)
        return {'type': COORD_TYPE_ETRSTM35FIN, 'N': ETRS['N'], 'E': ETRS['E']}
    else:
        return None



###########################################################################
# Function:  KKJxy_to_WGS84lalo
###########################################################################
# Input:     dictionary with ['P'] is KKJ Northing
#                            ['I'] in KKJ Easting
# Output:    dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
###########################################################################

def KKJxy_to_WGS84lalo(KKJin):  
  KKJz = KKJxy_to_KKJlalo(KKJin)
  WGS = KKJlalo_to_WGS84lalo(KKJz)

  return WGS



###########################################################################
# Function:  WGS84lalo_to_KKJxy
###########################################################################
# Input:     dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
# Output:    dictionary with ['P'] is KKJ Northing
#                            ['I'] in KKJ Eeasting
###########################################################################

def WGS84lalo_to_KKJxy(WGSin):
  KKJlalo = WGS84lalo_to_KKJlalo(WGSin);

  ZoneNumber = KKJ_Zone_Lo(KKJlalo['Lo'])
  KKJxy = KKJlalo_to_KKJxy(KKJlalo, ZoneNumber)

  return KKJxy



###########################################################################
# Function:  KKJlalo_to_WGS84lalo
###########################################################################

def KKJlalo_to_WGS84lalo(KKJ):
  La = KKJ['La']
  Lo = KKJ['Lo']

  dLa = math.radians(  0.124867E+01       + \
                      -0.269982E+00 * La + \
                     0.191330E+00 * Lo + \
                     0.356119E-02 * La * La + \
                     -0.122312E-02 * La * Lo + \
                     -0.335514E-03 * Lo * Lo ) / 3600.0
  dLo = math.radians( -0.286111E+02       + \
                    0.114183E+01 * La + \
                    -0.581428E+00 * Lo + \
                    -0.152421E-01 * La * La + \
                    0.118177E-01 * La * Lo + \
                    0.826646E-03 * Lo * Lo ) / 3600.0

  WGS = {}
  WGS['La'] = math.degrees(math.radians(KKJ['La']) + dLa)
  WGS['Lo'] = math.degrees(math.radians(KKJ['Lo']) + dLo)

  return WGS



###########################################################################
# Function:  WGS84lalo_to_KKJlalo
###########################################################################

def WGS84lalo_to_KKJlalo(WGS):
  La = WGS['La']
  Lo = WGS['Lo']

  dLa = math.radians( -0.124766E+01       + \
                    0.269941E+00 * La + \
                    -0.191342E+00 * Lo + \
                    -0.356086E-02 * La * La + \
                    0.122353E-02 * La * Lo + \
                    0.335456E-03 * Lo * Lo ) / 3600.0
  
  dLo = math.radians(  0.286008E+02       + \
                     -0.114139E+01 * La + \
                     0.581329E+00 * Lo + \
                     0.152376E-01 * La * La + \
                     -0.118166E-01 * La * Lo + \
                     -0.826201E-03 * Lo * Lo ) / 3600.0

  KKJ = {}
  KKJ['La'] = math.degrees(math.radians(WGS['La']) + dLa)
  KKJ['Lo'] = math.degrees(math.radians(WGS['Lo']) + dLo)

  return KKJ



###########################################################################
# Function:  KKJxy_to_KKJlalo
###########################################################################

def KKJxy_to_KKJlalo(KKJ):  
  #
  # Scan iteratively the target area, until find matching
  # KKJ coordinate value.  Area is defined with Hayford Ellipsoid.
  #  
  LALO = {}

  ZoneNumber = KKJ_Zone_I(KKJ['I'])
    
  MinLa = math.radians(59.0)
  MaxLa = math.radians(70.5)
  MinLo = math.radians(16.0)
  MaxLo = math.radians(35.5)

  i = 1
  while (i < 35):
    DeltaLa = MaxLa - MinLa
    DeltaLo = MaxLo - MinLo

    LALO['La'] = math.degrees(MinLa + 0.5 * DeltaLa)
    LALO['Lo'] = math.degrees(MinLo + 0.5 * DeltaLo)

    KKJt = KKJlalo_to_KKJxy(LALO, ZoneNumber)

    if (KKJt['P'] < KKJ['P']):
      MinLa = MinLa + 0.45 * DeltaLa
    else:
      MaxLa = MinLa + 0.55 * DeltaLa

    if (KKJt['I'] < KKJ['I']):
      MinLo = MinLo + 0.45 * DeltaLo
    else:
      MaxLo = MinLo + 0.55 * DeltaLo

    i = i + 1

  return LALO



###########################################################################
# Function:  KKJlalo_to_KKJxy
###########################################################################

def KKJlalo_to_KKJxy(INP, ZoneNumber):
  Lo = math.radians(INP['Lo']) - math.radians(KKJ_ZONE_INFO[ZoneNumber][0])

  a  = 6378388.0            # Hayford ellipsoid
  f  = 1/297.0

  b  = (1.0 - f) * a
  bb = b * b              
  c  = (a / b) * a        
  ee = (a * a - bb) / bb  
  n = (a - b)/(a + b)     
  nn = n * n              

  cosLa = math.cos(math.radians(INP['La']))

  NN = ee * cosLa * cosLa 

  LaF = math.atan(math.tan(math.radians(INP['La'])) / math.cos(Lo * math.sqrt(1 + NN)))

  cosLaF = math.cos(LaF)

  t   = (math.tan(Lo) * cosLaF) / math.sqrt(1 + ee * cosLaF * cosLaF)

  A   = a / ( 1 + n )

  A1  = A * (1 + nn / 4 + nn * nn / 64)

  A2  = A * 1.5 * n * (1 - nn / 8)

  A3  = A * 0.9375 * nn * (1 - nn / 4)

  A4  = A * 35/48.0 * nn * n

  OUT = {}
  OUT['P'] = A1 * LaF - \
        A2 * math.sin(2 * LaF) + \
            A3 * math.sin(4 * LaF) - \
                A4 * math.sin(6 * LaF)
  OUT['I'] = c * math.log(t + math.sqrt(1+t*t)) + \
        500000.0 + ZoneNumber * 1000000.0

  return OUT



###########################################################################
# Function:  KKJ_Zone_I
###########################################################################

def KKJ_Zone_I(KKJI):
  ZoneNumber = math.floor((KKJI/1000000.0))
  if ZoneNumber < 0 or ZoneNumber > 5:
      ZoneNumber = -1
      
  return ZoneNumber



###########################################################################
# Function:  KKJ_Zone_Lo
###########################################################################

def KKJ_Zone_Lo(KKJlo):
  # determine the zonenumber from KKJ easting
  # takes KKJ zone which has center meridian
  # longitude nearest (in math value) to
  # the given KKJ longitude
  ZoneNumber = 5
  while ZoneNumber >= 0:
    if math.fabs(KKJlo - KKJ_ZONE_INFO[ZoneNumber][0]) <= 1.5:
      break
    ZoneNumber = ZoneNumber - 1
            
  return ZoneNumber


###########################################################################
# Function:  KKJxy_ZoneShift
###########################################################################

def KKJxy_ZoneShift(KKJ, zone):
  kkjlalo = KKJxy_to_KKJlalo(KKJ)
  return KKJlalo_to_KKJxy(kkjlalo, zone)  


###########################################################################
# Function:  WGS84lalo_to_GoogleMapsXY
###########################################################################
# Input:     dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
#            google zoom factor (integer between 0 and 17, 0 for no zoom 
#                            and 17 for maximum zoom)
# Output:    dictionary with ['x'] is Google maps URL x parameter (tile number)
#                            ['y'] in Google maps URL y parameter (tile number)
###########################################################################

def WGS84lalo_to_GoogleMapsXY(WGSin, zoom):
  # Google maps maximum zoom factor (min = 0)
  MAXZOOM = 17.0

  worldwidth = math.pow(2.0, (MAXZOOM - zoom))
  x = WGSin['Lo'] + 180.0
  y = math.log(math.tan((math.pi / 4.0) + ((0.5 * math.pi * WGSin['La']) / 180.0))) / math.pi
  if y <-0.9999:
    y = -0.9999
  if y > 0.9999:
    y = 0.9999 
  y = (-90.0 * y + 90.0)
  
  out = {}
  out['x'] = (int) (math.floor((x / 360.0) * worldwidth))
  out['y'] = (int) (math.floor((y / 180.0) * worldwidth))

  return out



###########################################################################
# Function:  Str_to_CoordinateValue
###########################################################################
# Input:     string with a coordinate value in some WGS84-format
# 
# Output:    floating point value representing the given coordinate
#            Value returned is INVALID_COORDINATE if the given string cannot be
#            interpreted as WGS84 coordinate value. 
###########################################################################
INVALID_COORDINATE = -99999.99
def Str_to_CoordinateValue(WGSstr):
  
  # case 1: 61,27,4.96 (for 61 degrees, 27 minutes, 4.96 seconds)
  regexp1 = '^(?P<sig>-?)(?P<deg>\d+),(?P<min>\d+),(?P<sec>\d+(\.\d+)?)$'
  mo = re.match(regexp1, WGSstr)
  if (mo != None):
    value = string.atof( mo.group('deg') ) + string.atof( mo.group('min') ) / 60.0 + string.atof( mo.group('sec') ) / 3600.0
    if ( mo.group('sig') == '-' ):
      value = -value
    return value

  # case 2: 61,27.083 (for 61 degrees, 27.083 minutes)
  regexp2 = '^(?P<sig>-?)(?P<deg>\d+),(?P<min>\d+(\.\d+))?$'
  mo = re.match(regexp2, WGSstr)
  if (mo != None):
    value = string.atof( mo.group('deg') ) + string.atof( mo.group('min') ) / 60.0
    if ( mo.group('sig') == '-' ):
      value = -value
    return value

  # case 3: 61.451378 (for 61.451378 degrees)
  regexp3 = '^(?P<sig>-?)(?P<deg>\d+\.\d+)$'
  mo = re.match(regexp3, WGSstr)
  if (mo != None):
    value = string.atof( mo.group('deg') )
    if ( mo.group('sig') == '-' ):
        value = -value
    return value

  # case: other
  return INVALID_COORDINATE 



###########################################################################
# Function:  KKJxy_in_Finland
###########################################################################
# Input:     dictionary with ['P'] is KKJ Northing
#                            ['I'] in KKJ Eeasting
# Output:    truth value telling whether the given coordinate is 
#            _*_*_approximately_*_*_ in Finnish area. 
###########################################################################

FINLAND_AREA_ZONE3_BOXES = [ [6641806, 3069283, 6737806, 3161283], \
                             [6641806, 3161283, 7067806, 3755283], \
                             [7067806, 3253283, 7581806, 3691283], \
                             [7581806, 3413283, 7786097, 3595283], \
                             [7581806, 3219283, 7726097, 3413283]
                            ]

def KKJxy_in_Finland(KKJ):
  try:
    # Move the coordinates to zone 3
    lalo = KKJxy_to_KKJlalo(KKJ)
    xy = KKJlalo_to_KKJxy(lalo, 3)
 
    P = int(round(xy['P']))
    I = int(round(xy['I']))

    # go through the boxes and check if coordinates are inside one of them 
    for box in FINLAND_AREA_ZONE3_BOXES:
      if P >= box[0] and P <= box[2] and I >= box[1] and I <= box[3]: 
        return 1
  except:
    # If KKJ-conversion functions fail, assume it is not valid Finnish location
    return 0
     
  return 0



###########################################################################
# Constants used in calculations transforming between WGS84 and ETRS-TM35FIN 

Ca = 6378137.0
Cb = 6356752.314245
Cf = 1.0 / 298.257223563
Ck0 = 0.9996
Clo0 = math.radians(27.0)
CE0 = 500000.0
Cn = Cf / (2.0 - Cf)
CA1 = Ca / (1.0 + Cn) * (1.0 + math.pow(Cn, 2.0) / 4.0 + math.pow(Cn, 4.0) / 64.0)
Ce = math.sqrt(2.0 * Cf - math.pow(Cf, 2.0))
Ch1 = 1.0/2.0 * Cn - 2.0/3.0 * math.pow(Cn, 2.0) + 37.0/96.0 * math.pow(Cn, 3.0) - 1.0/360.0 * math.pow(Cn, 4.0)
Ch2 = 1.0/48.0 * math.pow(Cn, 2.0) + 1.0/15.0 * math.pow(Cn, 3.0) - 437.0/1440.0 * math.pow(Cn, 4.0)
Ch3 = 17.0/480.0 * math.pow(Cn, 3.0) - 37.0/840.0 * math.pow(Cn, 4.0)
Ch4 = 4397.0/161280.0 * math.pow(Cn, 4.0)
Ch1p = 1.0/2.0 * Cn - 2.0/3.0 * math.pow(Cn, 2.0) + 5.0/16.0 * math.pow(Cn, 3.0) + 41.0/180.0 * math.pow(Cn, 4.0)
Ch2p = 13.0/48.0 * math.pow(Cn, 2.0) - 3.0/5.0 * math.pow(Cn, 3.0) + 557.0/1440.0 * math.pow(Cn, 4.0)
Ch3p = 61.0/240.0 * math.pow(Cn, 3.0) - 103.0/140.0 * math.pow(Cn, 4.0)
Ch4p = 49561.0/161280.0 * math.pow(Cn, 4.0)

# inverse hyperbolic functions not found in Python math library 
# before Python version 2.6

def asinh(x):
  return math.log(x + math.sqrt(x * x + 1.0))

def atanh(x):
  return math.log((1.0 + x) / (1.0 - x)) / 2.0


###########################################################################
# Function:  ETRSTM35FINxy_to_WGS84lalo
###########################################################################
# Input:     dictionary with ['N'] is ETRS-TM35FIN Northing
#                            ['E'] in ETRS-TM35FIN Easting
# Output:    dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
###########################################################################

def ETRSTM35FINxy_to_WGS84lalo(ETRSin):  
  E = ETRSin['N'] / (CA1 * Ck0)
  nn = (ETRSin['E'] - CE0) / (CA1 * Ck0)
  
  E1p = Ch1 * math.sin(2.0 * E) * math.cosh(2.0 * nn)
  E2p = Ch2 * math.sin(4.0 * E) * math.cosh(4.0 * nn)
  E3p = Ch3 * math.sin(6.0 * E) * math.cosh(6.0 * nn)
  E4p = Ch4 * math.sin(8.0 * E) * math.cosh(8.0 * nn)
  nn1p = Ch1 * math.cos(2.0 * E) * math.sinh(2.0 * nn)
  nn2p = Ch2 * math.cos(4.0 * E) * math.sinh(4.0 * nn)
  nn3p = Ch3 * math.cos(6.0 * E) * math.sinh(6.0 * nn)
  nn4p = Ch4 * math.cos(8.0 * E) * math.sinh(8.0 * nn)
  Ep = E - E1p - E2p - E3p - E4p
  nnp = nn - nn1p - nn2p - nn3p - nn4p
  be = math.asin(math.sin(Ep) / math.cosh(nnp))
  
  Q = asinh(math.tan(be))
  Qp = Q + Ce * atanh(Ce * math.tanh(Q))
  Qp = Q + Ce * atanh(Ce * math.tanh(Qp))
  Qp = Q + Ce * atanh(Ce * math.tanh(Qp))
  Qp = Q + Ce * atanh(Ce * math.tanh(Qp))
  
  WGS = {}
  WGS['La'] = math.degrees(math.atan(math.sinh(Qp))) 
  WGS['Lo'] = math.degrees(Clo0 + math.asin(math.tanh(nnp) / math.cos(be)))
  
  return WGS



###########################################################################
# Function:  WGS84lalo_to_ETRSTM35FINxy
###########################################################################
# Input:     dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
# Output:    dictionary with ['N'] is ETRS-TM35FIN Northing
#                            ['E'] in ETRS-TM35FIN Eeasting
###########################################################################

def WGS84lalo_to_ETRSTM35FINxy(WGSin):
  la = math.radians(WGSin['La'])
  lo = math.radians(WGSin['Lo'])  
    
  Q = asinh(math.tan(la)) - Ce * atanh(Ce * math.sin(la))
  be = math.atan(math.sinh(Q))
  nnp = atanh(math.cos(be) * math.sin(lo - Clo0))
  Ep = math.asin(math.sin(be) * math.cosh(nnp))  
  E1 = Ch1p * math.sin(2.0 * Ep) * math.cosh(2.0 * nnp)
  E2 = Ch2p * math.sin(4.0 * Ep) * math.cosh(4.0 * nnp)
  E3 = Ch3p * math.sin(6.0 * Ep) * math.cosh(6.0 * nnp)
  E4 = Ch4p * math.sin(8.0 * Ep) * math.cosh(8.0 * nnp)
  nn1 = Ch1p * math.cos(2.0 * Ep) * math.sinh(2.0 * nnp)
  nn2 = Ch2p * math.cos(4.0 * Ep) * math.sinh(4.0 * nnp)
  nn3 = Ch3p * math.cos(6.0 * Ep) * math.sinh(6.0 * nnp)
  nn4 = Ch4p * math.cos(8.0 * Ep) * math.sinh(8.0 * nnp)
  E = Ep + E1 + E2 + E3 + E4
  nn = nnp + nn1 + nn2 + nn3 + nn4
  
  ETRS = {}
  ETRS['N'] = CA1 * E * Ck0
  ETRS['E'] = CA1 * nn * Ck0 + CE0

  return ETRS



###########################################################################
# Function:  KKJxy_to_ETRSTM35FINxy
###########################################################################
# Input:     dictionary with ['P'] is KKJ Northing
#                            ['I'] in KKJ Easting
# Output:    dictionary with ['N'] is ETRS-TM35FIN Northing
#                            ['E'] is ETRS-TM35FIN Easting
###########################################################################

def KKJxy_to_ETRSTM35FINxy(KKJin):
  return WGS84lalo_to_ETRSTM35FINxy(KKJxy_to_WGS84lalo(KKJin))  



###########################################################################
# Function:  ETRSTM35FINxy_to_KKJxy
###########################################################################
# Input:     dictionary with ['N'] is ETRS-TM35FIN Northing
#                            ['E'] is ETRS-TM35FIN Easting
# Output:    dictionary with ['P'] is KKJ Northing
#                            ['I'] in KKJ Easting
###########################################################################

def ETRSTM35FINxy_to_KKJxy(ETRSin):
  return WGS84lalo_to_KKJxy(ETRSTM35FINxy_to_WGS84lalo(ETRSin))  



###########################################################################
# Function:  WGS84distance
###########################################################################
# Input:     dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
#            dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
# Output:    distance between given points in meters
###########################################################################

def WGS84distance(WGSin1, WGSin2):
  dLon = math.radians(WGSin2['Lon'] - WGSin1['Lon'])
    
  U1 = math.atan((1.0 - Cf) * math.tan(math.radians(WGSin1['Lat'])))
  U2 = math.atan((1.0 - Cf) * math.tan(math.radians(WGSin2['Lat'])))
  sinU1 = math.sin(U1) 
  cosU1 = math.cos(U1)
  sinU2 = math.sin(U2)
  cosU2 = math.cos(U2)
  
  lam = dLon
  lamPrev = None
  limit = 100
  
  cosAlpha2 = 0.0
  sinSig = 0.0
  cosSig = 0.0
  cos2sigM = 0.0

  while ((not lamPrev or abs(lam - lamPrev) > 1e-12) and limit > 0):
    lamPrev = lam;
    sinLam = math.sin(lam)
    cosLam = math.cos(lam)
    
    sinSig = math.sqrt(math.pow(cosU2 * sinLam, 2.0) + math.pow(cosU1 * sinU2 - sinU1 * cosU2 * cosLam, 2.0))
    if sinSig == 0.0:
      return 0.0  # co-incident points
    cosSig = sinU1 * sinU2 + cosU1 * cosU2 * cosLam
    sig = math.atan2(sinSig, cosSig)
    sinAlpha = cosU1 * cosU2 * sinLam / sinSig
    cosAlpha2 = 1.0 - math.pow(sinAlpha, 2.0)

    if (cosAlpha2 == 0.0):
      cos2sigM = 0.0
    else:
      cos2sigM = cosSig - 2.0 * sinU1 * sinU2 / cosAlpha2

    C = Cf / 16.0 * cosAlpha2 * (4.0 + Cf * (4.0 - 3.0 * cosAlpha2))
    
    lam = dLon + (1.0 - C) * Cf * sinAlpha * (sig + C * sinSig * (cos2sigM + C * cosSig * (-1.0 + 2.0 * math.pow(cos2sigM, 2.0))))

  if limit == 0:
    return None 

  Cb2 = math.pow(Cb, 2.0)
  u2 = cosAlpha2 * (math.pow(Ca, 2.0) - Cb2) / Cb2
  A = 1.0 + u2 / 16384.0 * (4096.0 + u2 * (-768.0 + u2 * (320.0 - 175.0 * u2)))
  B = u2 / 1024.0 * (256.0 + u2 * (-128.0 + u2 * (74.0 - 47.0 * u2)))
  dsig = B * sinSig * (cos2sigM + 0.25 * B * (cosSig * (-1.0 + 2.0 * math.pow(cos2sigM, 2.0)) - \
         1.0 / 6.0 * B * cos2sigM * (-3.0 + 4.0 * math.pow(sinSig, 2.0)) * (-3.0 + 4.0 * math.pow(cos2sigM, 2.0))))
  
  return Cb * A * (sig - dsig)



###########################################################################
# Function:  WGS84bearing
###########################################################################
# Input:     dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
#            dictionary with ['La'] is latitude in degrees (WGS84)
#                            ['Lo'] is longitude in degrees (WGS84)
# Output:    tuple containing initial and final bearing (deg) from point 1 to 2 
###########################################################################

def WGS84bearing(WGSin1, WGSin2):
  dLon = math.radians(WGSin2['Lon'] - WGSin1['Lon'])
    
  U1 = math.atan((1.0 - Cf) * math.tan(math.radians(WGSin1['Lat'])))
  U2 = math.atan((1.0 - Cf) * math.tan(math.radians(WGSin2['Lat'])))
  sinU1 = math.sin(U1) 
  cosU1 = math.cos(U1)
  sinU2 = math.sin(U2)
  cosU2 = math.cos(U2)
  
  lam = dLon
  lamPrev = None
  limit = 100

  sinLam = 0.0
  cosLam = 0.0

  while ((not lamPrev or abs(lam - lamPrev) > 1e-12) and limit > 0):
    lamPrev = lam;
    sinLam = math.sin(lam)
    cosLam = math.cos(lam)
    
    sinSig = math.sqrt(math.pow(cosU2 * sinLam, 2.0) + math.pow(cosU1 * sinU2 - sinU1 * cosU2 * cosLam, 2.0))
    if sinSig == 0.0:
      return 0.0  # co-incident points
    cosSig = sinU1 * sinU2 + cosU1 * cosU2 * cosLam
    sig = math.atan2(sinSig, cosSig)
    sinAlpha = cosU1 * cosU2 * sinLam / sinSig
    cosAlpha2 = 1.0 - math.pow(sinAlpha, 2.0)

    if (cosAlpha2 == 0.0):
      cos2sigM = 0.0
    else:
      cos2sigM = cosSig - 2.0 * sinU1 * sinU2 / cosAlpha2

    C = Cf / 16.0 * cosAlpha2 * (4.0 + Cf * (4.0 - 3.0 * cosAlpha2))
    
    lam = dLon + (1.0 - C) * Cf * sinAlpha * (sig + C * sinSig * (cos2sigM + C * cosSig * (-1.0 + 2.0 * math.pow(cos2sigM, 2.0))))

  if limit == 0:
    return None 
  
  al1 = math.degrees(math.atan2(cosU2 * sinLam, cosU2 * sinU2 - sinU1 * cosU2 * cosLam))
  al2 = math.degrees(math.atan2(cosU1 * sinLam, -sinU1 * cosU2 + cosU1 * sinU2 * cosLam))
  
  return (al1, al2)