#! python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            fms_rka.py
#
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Not allowed to be used for mass downloading
#                  of maps!
#
#
HELP_TEXT = """
"""
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         3.0k
#
# Date:            03.11.2010
#
# Functions:       -
#                   
# Description:     Map service class for Retkikartta.fi.
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.1)
# 
#                  Python Imaging Library (tested with 1.1.5) 
#                  (www.pythonware.org/products/pil/)
#
# Version history: ** 18.02.2009 v3.0a (Olli Lammi) **
#                  First implementation.
#
#                  ** 19.02.2009 v3.0b (Olli Lammi) **
#                  Changes to the map construction mathematics and calibration.
#
#                  ** 23.03.2009 v3.0d (Olli Lammi) **
#                  Small change to service code.
#
#                  ** 26.03.2009 v3.0e (Olli Lammi) **
#                  Small change to service code.
#
#                  ** 27.09.2010 v3.0j (Olli Lammi) **
#                  Support for changed Retkikartta.fi platform.
#
#                  ** 03.11.2010 v3.0k (Olli Lammi) **
#                  Changed scale names to match scale name
#                  changes in the Retkikartta.fi.
#
###########################################################################

# Imports

import sys, os, string
import math, re, os.path
import time, random
import tempfile, shutil

import urllib2
import xml.dom.minidom

from PIL import Image  # Python Imaging Library 
                       # (http://www.pythonware.com/products/pil/)

import coordinates  # KKJ to and from WGS84 coordinate transforms

import fms

###########################################################################

# Constants

MOD_VERSION = 'v3.0k'
SERVICE_CODE = 'rka'

# constants for the picture and distance sizes and ratios
PIC_ORIG_SIZEX = 256
PIC_ORIG_SIZEY = 256
PIC_SIZEX = PIC_ORIG_SIZEX
PIC_SIZEY = PIC_ORIG_SIZEY


PIC_SCALES = { '1:5000': [5000, 1.411246], \
               '1:8000': [8000, 2.116577], \
               '1:16000': [16000, 4.234905], \
               '1:25000': [25000, 6.614597], \
               '1:50000': [50000, 13.229193], \
               '1:100000': [100000, 26.458386], \
               '1:500000': [500000, 132.291931], \
               '1:1000000': [1000000, 264.639239], \
               '1:3000000': [3000000, 794.034895] }

PIC_SCALE_DEFAULT = '1:16000'


###########################################################################

# Classes

class FMSRetkikarttaFi(fms.FetchMapService):
    def __init__(self):
        self.SERVICE_CODE = SERVICE_CODE
          
        self.PIC_SCALE_DEFAULT = PIC_SCALE_DEFAULT
        self.SetScale(self.PIC_SCALE_DEFAULT)
     
    
    ###########################################################################
    # Function:    GetVersion
    ###########################################################################
    # Description: Returns the module version string. 
    ###########################################################################    
    def GetVersion(self):
        return MOD_VERSION
    
    
    ###########################################################################
    # Function:    HandleParameter
    ###########################################################################
    # Description: Handles command line parameter. Returns 0, if the parameter
    #              was invalid. Returns 1 if the parameter was ok. 
    ###########################################################################  
    def HandleParameter(self, param):    
        return 0


    ###########################################################################
    # Function:    SetScale
    ###########################################################################
    # Description: Initializes the service and values to retrieve map parts
    #              in given scale. The parameter must be one of the values
    #              returned by GetScales.
    ###########################################################################
    def SetScale(self, scale):
        # determine picture scale and ratio
        try:
            self.PIC_SIZEX = PIC_SIZEX
            self.PIC_SIZEY = PIC_SIZEY
            self.SCALE = scale
            self.SCALEVAL = PIC_SCALES[scale][0]
            self.PIC_RATIO = PIC_SCALES[scale][1]
            self.MAP_SIZEX = self.PIC_SIZEX * self.PIC_RATIO
            self.MAP_SIZEY = self.PIC_SIZEY * self.PIC_RATIO
        except:
            return 0
        return 1


    ###########################################################################
    # Function:    GetScales
    ###########################################################################
    # Description: Returns list of scales that the service supports
    ###########################################################################    
    def GetScales(self):
        scales = []
        for sc in PIC_SCALES.keys():
            scales.append([PIC_SCALES[sc], sc])
        scales.sort()
        scales2 = []
        for sc in scales:
            scales2.append(sc[1])
        return scales2
    
    
    ###########################################################################
    # Function:    GetPicIndexes
    ###########################################################################
    # Description: Returns two (x,y) tuples with indexes for map image that has
    #              the given (mapx, mapy) point and the lower left xy
    #              coordinates of the indexed map picture.
    ###########################################################################
    def GetPicIndexes(self, mapx, mapy):
        # calculate the tile column and row
        indx = int((mapx + (0.25 * self.PIC_RATIO)) / (self.PIC_RATIO * PIC_ORIG_SIZEX))
        indy = int((mapy + (0.25 * self.PIC_RATIO)) / (self.PIC_RATIO * PIC_ORIG_SIZEY))
        fetchx = self.PIC_RATIO * indx * PIC_ORIG_SIZEX
        fetchy = self.PIC_RATIO * indy * PIC_ORIG_SIZEY
        
        return ((indx, indy), (fetchx, fetchy))

    
    ###########################################################################
    # Method:      FetchPic
    ###########################################################################
    # Description: Fetches one map tile from Retkikartta.fi.
    #              Result file will be fname. Tempdir is used as temp directory  
    #              while downloading.
    ###########################################################################
    def FetchPic(self, fetchr, fetchc, fname, tempdir):  
        url = 'http://ntile1.navici.com/t/retkikartta/maps/2393x' + str(self.PIC_RATIO) + 'x' + str(fetchc) + 'x' + str(fetchr) + '.png'

        USERAGENT = "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.0.4) Gecko/20060508 Firefox/1.5.0.4"
  
        res = ''
        try:
          req = urllib2.Request(url)
          req.add_header("User-Agent", USERAGENT)
  
          f = urllib2.urlopen(req)
          res = f.read()
          f.close()
        except:
          return 0
   
        tempfname = ''
        # create a temporary file name for the map
        try:
            (tempfd, tempfname) = tempfile.mkstemp('.png', 'tmpmap_', tempdir)
            os.close(tempfd)
        except:
            print 'ERROR: Cannot create temporary file for map image file'
            return 0

        try:
            f = open(tempfname,'wb')
            f.write(res)
            f.close()
        except:
            print "ERROR: Cannot write temporary map image file: " + tempfname
            return 0
    
        try:
            im = Image.open(tempfname)
            im.load()
    
            if im.size[0] != PIC_ORIG_SIZEX or im.size[1] != PIC_ORIG_SIZEY:
                print 'ERROR: Retrieved image is of invalid size (received: ' + `im.sizex` + 'x' + `im.sizey` + ', expected: ' + `PIC_ORIG_SIZEX` + 'x' + `PIC_ORIG_SIZEY` + ')'
                os.unlink(tempfname)
                return 0 
    
            im.save(tempfname)
        except Exception, e:
            print "ERROR: Cannot convert map image or rewrite temporary map image file: " + tempfname
            print e.__str__()
            os.unlink(tempfname)
            return 0
 
        # check if the map cache hash directory is created
        dname = os.path.dirname(fname)
        if not os.path.isdir(dname):
            try:
                os.makedirs(dname)
            except:
                print 'ERROR: Cannot create map image directory: ' + dname
                os.unlink(tempfname)
                return 0
    
        # move the temporary map file to the cache
        try:
            shutil.move(tempfname, fname)  
        except:
            print 'ERROR: Cannot move loaded temporary map image to map cache.'
            try:
                os.unlink(tempfname)
            except:
                return 0
            return 0
  
        return 1
         
         
    ###########################################################################
    # Method:    GetCacheName
    ###########################################################################
    # Description: Generates a name for the map image in cache according to
    #              location and other parameters.
    ###########################################################################
    def GetCacheName(self, x, y):
        fname = `self.SCALEVAL` + '_' + `int(self.PIC_SIZEX)` + '_' + `int(self.PIC_SIZEY)`
  
        # calculate location hash
        # convert to zone 3
        xy = coordinates.KKJxy_ZoneShift( {'P': y, 'I': x}, 3 )
        hashy = int(round(xy['P'])) 
        hashx = int(round(xy['I']))
        hash = `HashRound(hashy)` + '_' + `HashRound(hashx)`
        fname = fname + '/' + hash
  
        fname = fname + '/map_' + self.SERVICE_CODE + '_' + `self.SCALEVAL`  
        
        fname = fname + '_'+ repr(int(round(y))) + '_' + repr(int(round(x))) + '.png'

        return fname


    ###########################################################################
    # Method:      PostProcess
    ###########################################################################
    # Description: Post processes the entire constructed map image according
    #              to service requirements.
    ###########################################################################    
    def PostProcess(self, im):
        pass


    ###########################################################################
    # Method:      TranslateToServiceCoordinatesXY
    ###########################################################################
    # Description: Does service require coordinates to be transferred to YKJ?
    #              Returns: tuple (type, x, y) where
    #              type - type string of the coordinate system
    #              x - easting or lon coordinate 
    #              y - northing or lat coordinate
    ###########################################################################       
    def TranslateToServiceCoordinatesXY(self, type, x, y):
        if (type != coordinates.COORD_TYPE_YKJ):
            temp = coordinates.Translate({'type': type, 'N': y, 'E': x}, \
                                         coordinates.COORD_TYPE_YKJ)
            return (temp['type'], temp['E'], temp['N'])
        return (type, x, y)

    
    ###########################################################################
    # Method:      GetHelp
    ###########################################################################
    # Description: Get service module specific help string.
    ###########################################################################    
    def GetHelp(self):
        return HELP_TEXT


    ###########################################################################
    # Method:      GetServiceDelayLimits
    ###########################################################################
    # Description: Get service specific delay times as tuple (min, max) in sec.
    ###########################################################################               
    def GetServiceDelayLimits(self):
        return (2, 6)
        
        
###########################################################################

# Functions

def HashRound(x):
    return ((x / 10000) / 5) * 5 


